# based on Andrey Vasnetsov code: https://gist.github.com/generall/68fddb87ae1845d6f54c958ed3d0addb

import numpy as np

from collections import defaultdict
from copy import deepcopy
from .utils import ft_ngram_hashes


def count_buckets(ft, words, new_ngrams_size):
    new_to_old_buckets = defaultdict(set)
    old_hash_count = defaultdict(int)
    for word in words:
        old_hashes = ft_ngram_hashes(word=word, minn=ft.min_n, maxn=ft.max_n, num_buckets=ft.bucket)
        new_hashes = ft_ngram_hashes(word=word, minn=ft.min_n, maxn=ft.max_n, num_buckets=new_ngrams_size)
        for old_hash in old_hashes:
            old_hash_count[old_hash] += 1  # calculate frequency of ngrams for proper weighting

        for old_hash, new_hash in zip(old_hashes, new_hashes):
            new_to_old_buckets[new_hash].add(old_hash)
    return new_to_old_buckets, old_hash_count


def fasttext_like_init(n, dim=300, random_state=42):
    rand_obj = np.random
    rand_obj.seed(random_state)
    lo, hi = -1.0 / dim, 1.0 / dim
    new_ngrams = rand_obj.uniform(lo, hi, (n, dim)).astype(np.float32)
    return new_ngrams


def prune_ngrams(ft, new_ngrams_size, random_state=42):
    """ Reduce the size of fasttext ngrams matrix by collapsing some hashes together """
    new_to_old_buckets, old_hash_count = count_buckets(
        ft, ft.key_to_index.keys(), new_ngrams_size
    )

    # initialize new buckets just like in fasttext
    new_ngrams = fasttext_like_init(n=new_ngrams_size, dim=ft.vectors_ngrams.shape[1], random_state=random_state)

    # fill new buckets with the weighted average old vectors
    for new_hash, old_buckets in new_to_old_buckets.items():
        total_sum = sum(old_hash_count[old_hash] for old_hash in old_buckets)
        new_vector = np.zeros(ft.vector_size, dtype=np.float32)
        for old_hash in old_buckets:
            weight = old_hash_count[old_hash] / total_sum
            new_vector += ft.vectors_ngrams[old_hash] * weight
        new_ngrams[new_hash] = new_vector

    return new_ngrams


def prune_vocab(ft, new_vocab_size=1_000):
    sorted_vocab = sorted(ft.key_to_index.items(), key=lambda x: ft.get_vecattr(x[0], 'count'), reverse=True)
    top_vocab_list = deepcopy(sorted_vocab[:new_vocab_size])
    top_vector_ids = []
    for new_index, (word, idx) in enumerate(top_vocab_list):
        top_vector_ids.append(idx)
    top_vectors = ft.vectors[top_vector_ids, :]
    return dict(top_vocab_list), top_vectors


class RowSparseMatrix:
    """ This data structure returns 0 for empty rows """
    def __init__(self, mapping, compressed, nrows, ncols, dtype):
        self.mapping = mapping
        self.compressed = compressed
        self.nrows = nrows
        self.ncols = ncols
        self.dtype = dtype

    def __getitem__(self, item):
        if item not in self.mapping:
            return np.zeros(self.ncols, dtype=self.dtype)
        return self.compressed[self.mapping[item]]

    def unpack(self):
        result = np.zeros(shape=self.shape, dtype=self.dtype)
        for old_index, new_index in self.mapping.items():
            result[old_index] = self[new_index].astype(self.dtype)
        return result

    def __add__(self, other):
        return self.unpack() + other

    def __sub__(self, other):
        return self.unpack() - other

    def __mul__(self, other):
        return self.unpack() * other

    def __truediv__(self, other):
        return self.unpack() / other

    def __pow__(self, other):
        return self.unpack() ** other

    def sqrt(self):
        return self ** 0.5

    @property
    def shape(self):
        return self.nrows, self.ncols

    @classmethod
    def from_big(cls, row_indices, big_matrix, dtype=np.float32):
        nrows, ncols = big_matrix.shape
        dtype = dtype
        compressed = big_matrix[row_indices].astype(dtype)
        mapping = {old_id: new_id for new_id, old_id in enumerate(row_indices)}
        return cls(mapping, compressed, nrows, ncols, dtype)

    @classmethod
    def from_small(cls, row_indices, small_matrix, nrows, dtype=np.float32):
        nrows, ncols = nrows, small_matrix.shape[1]
        dtype = dtype
        compressed = small_matrix
        mapping = {old_id: new_id for new_id, old_id in enumerate(row_indices)}
        return cls(mapping, compressed, nrows, ncols, dtype)



import gensim
import logging
import numpy as np

from .quantization import quantize
from .decomposition import DecomposedMatrix
from .prune import prune_ngrams, prune_vocab, count_buckets, RowSparseMatrix


logger = logging.getLogger(__name__)


def make_new_fasttext_model(
        ft,
        new_vectors,
        new_vectors_ngrams,
        new_vocab=None,
        cls=None,
):
    cls = cls or gensim.models.fasttext.FastTextKeyedVectors
    new_ft = cls(
        vector_size=ft.vector_size,
        min_n=ft.min_n,
        max_n=ft.max_n,
        bucket=new_vectors_ngrams.shape[0],
    )
    new_ft.vectors_vocab = None  # if we don't fine tune the model we don't need these vectors
    new_ft.vectors = new_vectors  # quantized vectors top_vectors
    if new_vocab is None:
        new_ft.key_to_index = ft.key_to_index
    else:
        new_ft.key_to_index = new_vocab
    new_ft.vectors_ngrams = new_vectors_ngrams
    if hasattr(new_ft, 'update_index2word'):
        new_ft.update_index2word()
    return new_ft


def quantize_ft(ft, qdim=100, centroids=255, sample=None):
    logger.info('quantizing vectors...')
    new_vectors = quantize(ft.vectors, qdim=qdim, centroids=centroids, verbose=True, sample=sample)
    logger.info('quantizing ngrams...')
    new_vectors_ngrams = quantize(ft.vectors_ngrams, qdim=qdim, centroids=centroids, verbose=True, sample=sample)

    return make_new_fasttext_model(ft, new_vectors=new_vectors, new_vectors_ngrams=new_vectors_ngrams)


def svd_ft(ft, n_components=30, fp16=True):
    logger.info('compressing vectors...')
    new_vectors = DecomposedMatrix.compress(ft.vectors, n_components=n_components, fp16=fp16)
    logger.info('compressing ngrams...')
    new_vectors_ngrams = DecomposedMatrix.compress(ft.vectors_ngrams, n_components=n_components, fp16=fp16)

    return make_new_fasttext_model(ft, new_vectors=new_vectors, new_vectors_ngrams=new_vectors_ngrams)


def prune_ft(ft, new_vocab_size=1_000, new_ngrams_size=20_000, fp16=True):
    logger.info('compressing vectors...')
    top_vocab, top_vectors = prune_vocab(ft, new_vocab_size=new_vocab_size)
    logger.info('compressing ngrams...')
    new_ngrams = prune_ngrams(ft, new_ngrams_size)
    if fp16:
        top_vectors = top_vectors.astype(np.float16)
        new_ngrams = new_ngrams.astype(np.float16)
    return make_new_fasttext_model(
        ft,
        new_vectors=top_vectors,
        new_vectors_ngrams=new_ngrams,
        new_vocab=top_vocab,
    )


def prune_ft_freq(
        ft,
        new_vocab_size=20_000,
        new_ngrams_size=100_000,
        fp16=True,
        pq=True,
        qdim=100,
        centroids=255,
        prune_by_norm=True,
        norm_power=1,
):
    if prune_by_norm:
        ngram_norms = np.linalg.norm(ft.vectors_ngrams, axis=-1)
        scorer = lambda id, count: count * (ngram_norms[id] ** norm_power)
    else:
        scorer = lambda id, count: count

    logger.info('quantizing ngrams...')
    new_to_old_buckets, old_hash_count = count_buckets(
        ft, list(ft.key_to_index.keys()), new_ngrams_size=new_ngrams_size,
    )
    logger.info('old ngrams in use: {}'.format(len(old_hash_count)))
    id_and_count = sorted(old_hash_count.items(), key=lambda x: scorer(*x), reverse=True)
    ids = [x[0] for x in id_and_count[:new_ngrams_size]]
    top_ngram_vecs = ft.vectors_ngrams[ids]
    if pq and len(top_ngram_vecs) > 0:
        top_ngram_vecs = quantize(top_ngram_vecs, qdim=qdim, centroids=centroids)
    elif fp16:
        top_ngram_vecs = top_ngram_vecs.astype(np.float16)
    rsm = RowSparseMatrix.from_small(ids, top_ngram_vecs, nrows=ft.vectors_ngrams.shape[0])

    logger.info('quantizing vectors...')
    top_voc, top_vec = prune_vocab(ft, new_vocab_size=new_vocab_size)
    if pq and len(top_vec) > 0:
        top_vec = quantize(top_vec, qdim=qdim, centroids=centroids)
    elif fp16:
        top_vec = top_vec.astype(np.float16)

    return make_new_fasttext_model(ft, top_vec, rsm, new_vocab=top_voc)



import numpy as np
import sys
from types import ModuleType, FunctionType
from gc import get_referents

try:
    from gensim.models.utils_any2vec import ft_ngram_hashes
except ImportError:
    from gensim.models.fasttext import ft_ngram_hashes

# Custom objects know their class.
# Function objects seem to know way too much, including modules.
# Exclude modules as well.
BLACKLIST = type, ModuleType, FunctionType


def getsize(obj):
    """ sum size of object & members. """
    if isinstance(obj, BLACKLIST):
        raise TypeError('getsize() does not take argument of type: '+ str(type(obj)))
    seen_ids = set()
    size = 0
    objects = [obj]
    while objects:
        need_referents = []
        for obj in objects:
            if not isinstance(obj, BLACKLIST) and id(obj) not in seen_ids:
                seen_ids.add(id(obj))
                if isinstance(obj, np.ndarray):
                    size += obj.nbytes
                else:
                    size += sys.getsizeof(obj)
                need_referents.append(obj)
        objects = get_referents(*need_referents)
    return size


def mb(obj):
    return getsize(obj) / 2**20



from compress_fasttext import compress, decomposition, evaluation, navec_like, prune, quantization, utils
from compress_fasttext import models
from compress_fasttext.compress import make_new_fasttext_model, quantize_ft, svd_ft, prune_ft, prune_ft_freq



import numpy as np
import warnings

from gensim.models.fasttext import FastTextKeyedVectors
from .utils import ft_ngram_hashes


from compress_fasttext.compress import make_new_fasttext_model


EPSILON = 1e-24


class CompressedFastTextKeyedVectors(FastTextKeyedVectors):
    """ This class extends FastTextKeyedVectors by fixing several issues:
        - index2word of a freshly created model is initialized from its vocab
        - the model does not keep heavy and useless vectors_ngrams_norm
        - word_vec() method with use_norm applies normalization in the right place
    """
    def __init__(self, *args, **kwargs):
        super(CompressedFastTextKeyedVectors, self).__init__(*args, **kwargs)
        self.update_index2word()

    @classmethod
    def load(cls, *args, **kwargs):
        try:
            loaded = super(CompressedFastTextKeyedVectors, cls).load(*args, **kwargs)
        except Exception as e:
            warnings.warn(
                'You seem to be loading an old model (compressed with gensim<4.0.0, compress-fasttext<0.1.0) '
                'within an new environment (gensim>=4.0.0, compress-fasttext>=0.1.0), which may not work. '
                'Please downgrade the packages or re-compress the model within the new environment.'
            )
            raise e
        # print(loaded.__dict__)
        return make_new_fasttext_model(
            loaded,
            new_vectors=loaded.vectors,
            new_vectors_ngrams=loaded.vectors_ngrams,
            new_vocab=loaded.key_to_index,
            cls=cls,
        )

    def update_index2word(self):
        if not self.index_to_key:
            inverse_index = {value: key for key, value in self.key_to_index.items()}
            self.index_to_key = [inverse_index.get(i) for i in range(len(self.key_to_index))]

    def word_vec(self, word, use_norm=False):
        """Get `word` representations in vector space, as a 1D numpy array.

        Parameters
        ----------
        word : str
            Input word
        use_norm : bool, optional
            If True - resulting vector will be L2-normalized (unit euclidean length).

        Returns
        -------
        numpy.ndarray
            Vector representation of `word`.

        Raises
        ------
        KeyError
            If word and all ngrams not in vocabulary.

        """
        if word in self.key_to_index:
            return super(FastTextKeyedVectors, self).word_vec(word, use_norm)
        elif self.bucket == 0:
            raise KeyError('cannot calculate vector for OOV word without ngrams')
        else:
            word_vec = np.zeros(self.vectors_ngrams.shape[1], dtype=np.float32)
            ngram_hashes = ft_ngram_hashes(word=word, minn=self.min_n, maxn=self.max_n, num_buckets=self.bucket)
            if len(ngram_hashes) == 0:
                return word_vec
            for nh in ngram_hashes:
                word_vec += self.vectors_ngrams[nh]
            result = word_vec / len(ngram_hashes)
            if use_norm:
                result /= np.sqrt(max(sum(result ** 2), EPSILON))
            return result

    def fill_norms(self, force=False):
        """
        Ensure per-vector norms are available.

        Any code which modifies vectors should ensure the accompanying norms are
        either recalculated or 'None', to trigger a full recalculation later on-request.

        """
        if self.norms is None or force:
            # self.norms = np.linalg.norm(self.vectors, axis=1)
            self.norms = np.stack([sum(self.vectors[i]**2) ** 0.5 for i in range(len(self.vectors))])

    def init_sims(self, replace=False):
        """Precompute L2-normalized vectors.

        Parameters
        ----------
        replace : bool, optional
            If True - forget the original vectors and only keep the normalized ones = saves lots of memory!
        """
        super(FastTextKeyedVectors, self).init_sims(replace)
        # todo: make self.vectors_norm a view over self.vectors, to avoid decompression
        # do NOT calculate vectors_ngrams_norm; using them is a mistake
        self.vectors_ngrams_norm = None



import numpy as np

try:
    from sklearn.decomposition import TruncatedSVD
except ImportError:
    # SVD is not the core feature of this library, so we keep this dependency optional
    TruncatedSVD = None


class DecomposedMatrix:
    def __init__(self, compressed, components):
        self.compressed = compressed
        self.components = components

    def __getitem__(self, item):
        return np.dot(self.compressed[item], self.components)

    @property
    def shape(self):
        return self.compressed.shape[0], self.components.shape[1]

    @classmethod
    def compress(cls, data, n_components=30, fp16=True):
        if not TruncatedSVD:
            raise ImportError('You need to install the `scikit-learn` package to perform matrix decomposition')
        model = TruncatedSVD(n_components=n_components)
        compressed = model.fit_transform(data)
        if fp16:
            compressed = compressed.astype(np.float16)
        return cls(compressed, model.components_)



import numpy as np


class Record(object):
    __attributes__ = []

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and all(
                (getattr(self, _) == getattr(other, _))
                for _ in self.__attributes__
            )
        )

    def __ne__(self, other):
        return not self == other

    def __iter__(self):
        return (getattr(self, _) for _ in self.__attributes__)

    def __hash__(self):
        return hash(tuple(self))

    def __repr__(self):
        name = self.__class__.__name__
        args = ', '.join(
            '{key}={value!r}'.format(
                key=_,
                value=getattr(self, _)
            )
            for _ in self.__attributes__
        )
        return '{name}({args})'.format(
            name=name,
            args=args
        )

    def _repr_pretty_(self, printer, cycle):
        name = self.__class__.__name__
        if cycle:
            printer.text('{name}(...)'.format(name=name))
        else:
            printer.text('{name}('.format(name=name))
            keys = self.__attributes__
            size = len(keys)
            if size:
                with printer.indent(4):
                    printer.break_()
                    for index, key in enumerate(keys):
                        printer.text(key + '=')
                        value = getattr(self, key)
                        printer.pretty(value)
                        if index < size - 1:
                            printer.text(',')
                            printer.break_()
                printer.break_()
            printer.text(')')


class PQ(Record):
    __attributes__ = ['vectors', 'dim', 'qdim', 'centroids', 'indexes', 'codes']

    def __init__(self, vectors, dim, qdim, centroids, indexes, codes):
        self.vectors = vectors
        self.dim = dim
        self.qdim = qdim
        self.centroids = centroids
        self.indexes = indexes
        self.codes = codes
        self.qdims = np.arange(self.qdim)

    def __getitem__(self, id):
        indexes = self.indexes[id]
        parts = self.codes[self.qdims, indexes]
        if len(parts.shape) == 2:
            return parts.reshape(self.dim)
        else:  # the index has been list or slice
            return parts.reshape(parts.shape[0], self.dim)

    def __add__(self, other):
        return self.unpack() + other

    def __sub__(self, other):
        return self.unpack() - other

    def __mul__(self, other):
        return self.unpack() * other

    def __truediv__(self, other):
        return self.unpack() / other

    def __pow__(self, other):
        return self.unpack() ** other

    def sqrt(self):
        return self ** 0.5

    def unpack(self):
        parts = self.codes[self.qdims, self.indexes]
        return parts.reshape(self.vectors, self.dim)

    def sampled(self, ids):
        vectors = len(ids)
        indexes = self.indexes[ids]
        return PQ(
            vectors, self.dim, self.qdim, self.centroids,
            indexes, self.codes
        )

    @property
    def shape(self):
        return self.vectors, self.dim

    @property
    def as_bytes(self):
        meta = self.vectors, self.dim, self.qdim, self.centroids
        meta = np.array(meta).astype(np.uint32).tobytes()
        indexes = self.indexes.astype(self.index_type(self.centroids)).tobytes()
        codes = self.codes.astype(np.float32).tobytes()
        return meta + indexes + codes

    @classmethod
    def from_file(cls, file):
        buffer = file.read(4 * 4)
        vectors, dim, qdim, centroids = np.frombuffer(buffer, np.uint32)
        buffer = file.read(vectors * qdim)
        indexes = np.frombuffer(buffer, cls.index_type(centroids)).reshape(vectors, qdim)
        buffer = file.read()
        codes = np.frombuffer(buffer, np.float32).reshape(qdim, centroids, -1)
        return cls(vectors, dim, qdim, centroids, indexes, codes)

    @classmethod
    def index_type(cls, centroids):
        return np.uint8 if centroids <= 255 else np.uint16

    def __len__(self):
        return self.vectors



import logging
import numpy as np

try:
    import pqkmeans
except ImportError:
    pqkmeans = None

from .navec_like import PQ


def quantize(matrix, qdim, centroids, sample=None, iterations=5, verbose=False):
    if not pqkmeans:
        raise ImportError('You need to install the `pqkmeans` package to perform quantization')
    encoder = pqkmeans.encoder.PQEncoder(
        iteration=iterations,
        num_subdim=qdim,
        Ks=centroids
    )

    matrix = np.array(matrix)
    vectors, dim = matrix.shape
    if sample is None:
        selection = matrix
    else:
        indexes = np.random.randint(vectors, size=sample)
        selection = matrix[indexes]

    encoder.fit(selection)
    indexes = encoder.transform(matrix).astype(PQ.index_type(centroids))
    codes = encoder.codewords

    if verbose:
        mb = 1024**2
        old_size = matrix.nbytes / mb
        new_size1 = codes.nbytes / mb
        new_size2 = indexes.nbytes / mb
        logging.info('Matrix size went from {:.1f} mb to {:.1f} + {:.1f} mb ({:.1%})'.format(
            old_size, new_size1, new_size2, (new_size1 + new_size2) / old_size
        ))

    return PQ(vectors, dim, qdim, centroids, indexes, codes)



import os
import re
import numpy as np

from collections import Counter


def vocabulary_from_files(path):
    wc = Counter()
    for fn in os.listdir(path):
        if not fn.endswith('txt'):
            continue
        with open(path + fn, 'r', encoding='utf-8') as f:
            for line in f.readlines():
                for w in re.sub('[^а-яёa-z]', ' ', line.lower()).split():
                    w2 = w.strip().replace('ё', 'е')
                    if len(w2) > 1:  # one-letter words are not included in the model I compress, so I skip them
                        wc[w2] += 1
    return wc


def cosine(x, y, eps=1e-10):
    x = x.astype(np.float32)
    y = y.astype(np.float32)
    den = (np.dot(x,x) * np.dot(y,y))**0.5
    return np.dot(x, y) / (den + eps)


def vecs_similarity(old_vecs, new_vecs, corpus):
    return np.mean([cosine(old_vecs[w], new_vecs[w]) for w in corpus])


def make_evaluator(baseline_model, path):
    vocab = vocabulary_from_files(path)
    return lambda model: vecs_similarity(baseline_model, model, vocab.keys())



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="compress-fasttext",
    version="0.1.1",
    author="David Dale",
    author_email="dale.david@mail.ru",
    description="A set of tools to compress gensim fasttext models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/avidale/compress-fasttext",
    packages=['compress_fasttext'],
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'gensim>=4.0.0',
        'numpy',
    ],
    extras_require={
        'full': [
            'scikit-learn',
            'pqkmeans',
        ],
    }
)



from typing import List

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


def tokenize(text) -> List[str]:
    """ Keep only alphanumeric character and split the text by space"""
    return ''.join(x for x in text if x.isalnum() or x.isspace()).split() or [' ']


class FastTextTransformer(BaseEstimator, TransformerMixin):
    """ Convert texts into their mean fastText vectors """
    def __init__(self, model, tokenizer=tokenize, normalize=True):
        self.model = model
        self.tokenizer = tokenizer
        self.normalize = normalize

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        result = np.stack([
            np.mean([self.model[w] for w in self.tokenizer(text)], 0)
            for text in X
        ])
        if self.normalize:
            norm = np.maximum(1e-10, (result ** 2).sum(1, keepdims=True) ** 0.5)
            result /= norm
        return result
